// Copyright (C) 2010 - DIGITEO - Michael Baudin

// Updates the .xml files by deleting existing files and 
// creating them again from the .sci with help_from_sci.

function helpupdate ( funarray , helpdir , macrosdir , demosdir , modulename )
  // Update the help and the demos from the .sci files.
  //
  // Calling Sequence
  //   updatehelp ( funarray , helpdir , macrosdir , demosdir )
  //
  // Parameters
  //   funarray : column matrix of strings. The list of functions to update
  //   helpdir : the help directory
  //   macrosdir : the macros directory
  //   demosdir : the demonstration directory
  //   modulename : the name of the module
  //
  // Description
  //   Update the .xml help files and the demos scripts
  //   from the macros corresponding to the function array
  //   of strings funarray.
  //   The existing .xml files in the help dir which 
  //   correspond to file in the funarray are deleted (Caution !).
  //   Generates the .xml and the .sce files from the help_from_sci function.
  //   Generates a demonstration gateway.
  //
  //   If demosdir is an empty matrix, do not generate the demonstrations.
  //   Last update : 29/07/2010.
  //
  // Author
  //   2010, Michael Baudin
  
  if ( fileinfo ( helpdir ) == [] ) then
    error(sprintf(gettext("%s: Wrong help directory: %s does not exist.\n"),"updatehelp",helpdir));
  end
  if ( fileinfo ( macrosdir ) == [] ) then
    error(sprintf(gettext("%s: Wrong macros directory: %s does not exist.\n"),"updatehelp",macrosdir));
  end
  if ( demosdir <> [] ) then
    if ( fileinfo ( demosdir ) == [] ) then
      error(sprintf(gettext("%s: Wrong demos directory: %s does not exist.\n"),"updatehelp",demosdir));
    end
  end
  
  //
  // 1. Delete the existing .xml help files in the helpdir
  // which correspond to given function names.
  flist = ls(helpdir)';
  for f = flist
    isxml = regexp(f,"/(.*).xml/");
    kf = find(funarray==basename(f))
    if ( isxml <> [] & kf <> [] ) then
      xmlfile = fullfile ( helpdir , f )
      mprintf("Deleting %s\n",xmlfile);
      r = deletefile(xmlfile);
      if ( ~r ) then
        error(sprintf(gettext("%s: Unable to delete xml file: %s\n"),"updatehelp",xmlfile));
      end
    end
  end
  //
  // 2. Generate each .xml and each .sce from the .sci
  flist = ls(macrosdir)';
  for f = flist
    issci = regexp(f,"/(.*).sci/");
    kf = find(funarray==basename(f))
    if ( issci <> [] & kf <> [] ) then
      scifile = fullfile ( macrosdir , f )
      mprintf("Processing %s\n",scifile);
      funname = funarray(kf)
      [helptxt,demotxt]= help_from_sci (scifile)
      xmlfile = fullfile ( helpdir , funname + ".xml" )
      mprintf("  Writing xml %s\n",xmlfile);
      r = mputl ( helptxt , xmlfile )
      if ( ~r ) then
        error(sprintf(gettext("%s: Unable to write xml file: %s\n"),"updatehelp",xmlfile));
      end
      if ( demosdir <> [] ) then
        demofile = fullfile ( demosdir , funname + ".sce" )
        mprintf("  Writing demo %s\n",demofile);
        r = mputl ( demotxt , demofile )
        if ( ~r ) then
          error(sprintf(gettext("%s: Unable to write demo file: %s\n"),"updatehelp",demofile));
        end
      end
    end
  end
  //
  // 2.1 Update the demo script to include the editor command.
  // This lets the user see the script.
  //
  if ( demosdir <> [] ) then
    flist = ls(macrosdir)';
    for f = flist
      issci = regexp(f,"/(.*).sci/");
      kf = find(funarray==basename(f))
      if ( issci <> [] & kf <> [] ) then
        funname = funarray(kf)
        demofile = fullfile ( demosdir , funname + ".sce" )
        mprintf("Updating %s\n",demofile);
        [fd,err]=mopen(demofile ,"a")
        if ( err <> 0 ) then
          error(sprintf(gettext("%s: Unable to open demo file %s for update\n"),"updatehelp",demofile));
        end
        mfprintf(fd,"//\n");
        mfprintf(fd,"// Load this script into the editor\n");
        mfprintf(fd,"//\n");
        mfprintf(fd,"filename = ""%s"";\n",funname + ".sce");
        mfprintf(fd,"dname = get_absolute_file_path(filename);\n");
        mfprintf(fd,"editor ( dname + filename );\n");
        err=mclose(fd)
        if ( err <> 0 ) then
          error(sprintf(gettext("%s: Unable to close demo file %s for update\n"),"updatehelp",demofile));
        end
      end
    end
  end
  //
  // 3. Generates the Demonstration gateway
  // Include all .sce files in the gateway, including 
  // handcrafted scripts which might have been written
  // by the user (and not generated by this function).
  // To make so, use the list of .sce files in the demo dir and 
  /// not simply the funarray.
  if ( demosdir <> [] ) then
    template = []
    template ($+1) = "// This file is released into the public domain"
    w=getdate()
    strdate = sprintf("%d/%d/%d - %d:%d:%d",w(1),w(2),w(6),w(7),w(8),w(9))
    template ($+1) = "// This help file was generated using helpupdate at " + strdate
    template ($+1) = "demopath = get_absolute_file_path(""" + modulename + ".dem.gateway.sce"");"
    template ($+1) = "subdemolist = ["
    flist = ls(demosdir)';
    for f = flist
      issce = regexp(f,"/(.*).sce/");
      isgateway = regexp(f,"/(.*).gateway.sce/");
      if ( issce <> [] & isgateway == [] ) then
        flen = length(f)
        funname = part(f,[1:flen-4])
        template($+1) = """" + funname + """, """ + funname + ".sce""; .."
      end
    end
    //
    template ($+1) = "];"
    template ($+1) = "subdemolist(:,2) = demopath + subdemolist(:,2)"
    gatefile = fullfile ( demosdir , modulename+".dem.gateway.sce" )
    mprintf("Writing demo Gateway %s\n",gatefile);
    r = mputl ( template , gatefile )
    if ( ~r ) then
      error(sprintf(gettext("%s: Unable to write demo file: %s\n"),"updatehelp",demofile));
    end
  end
endfunction
//
cwd = get_absolute_file_path("update_help.sce");
mprintf("Working dir = %s\n",cwd);
//
// Generate the library help
mprintf("Updating fmincon\n");
helpdir = cwd;
funmat = [
  "fmincon"
  "optimget"
  "optimset"
  ];
macrosdir = helpdir +"../../macros/fmincon";
demosdir = helpdir +"../../demos";
modulename = "fmincon";
helpupdate ( funmat , helpdir , macrosdir , demosdir , modulename )

//
// Generate the API functions help.
// Do not generate the demos for these functions.
mprintf("Updating apifun\n");
helpdir = fullfile(cwd,"apifun");
funmat = [
  "fmincon_checkrhs"
  "fmincon_checklhs"
  "fmincon_checktype"
  "fmincon_checkoption"
  "fmincon_checkdims"
  "fmincon_checksquare"
  "fmincon_checkvector"
  "fmincon_checkvecrow"
  "fmincon_checkveccol"
  "fmincon_checkscalar"
  "fmincon_checkgreq"
  "fmincon_checkloweq"
  "fmincon_checkrange"
  ];
macrosdir = cwd +"../../macros/apifun";
demosdir = [];
modulename = "fmincon";
helpupdate ( funmat , helpdir , macrosdir , demosdir , modulename )


